#include <stdlib.h>
#include <android/bitmap.h>

#include "ComicFinder.hpp"

#ifdef ENABLE_NEON
#include <arm_neon.h>
#endif

JNIEXPORT bool JNICALL Java_com_example_comicfinder_MainActivity_processImage
    (JNIEnv *env, jobject me, jbyteArray src, jobject dst, jint width, jint height)
{
    int ret;
    unsigned char *p_dst;
    if ((ret = AndroidBitmap_lockPixels(env, dst, reinterpret_cast<void **>(&p_dst))) < 0) {
        return false;
    }

    unsigned char *p_src = reinterpret_cast<unsigned char *>(env->GetPrimitiveArrayCritical(src, NULL));
    if (NULL == p_src) {
        AndroidBitmap_unlockPixels(env, dst);
        return false;
    }

    const int threshold_black_and_gray = 60;
    const int threshold_gray_and_white = 120;
    const int threshold_edge = 32;

    int lum, fx, fy;
    unsigned char line_buf[(width+2)*3];
    unsigned char *line0 = line_buf;
    unsigned char *line1 = line0 + width+2;
    unsigned char *line2 = line1 + width+2;
    memcpy(line1+1, p_src, width);
    line1[0] = line1[1];
    line1[width+1] = line1[width];
    memcpy(line2+1, p_src+width, width);
    line2[0] = line2[1];
    line2[width+1] = line2[width];
    memcpy(line0, line1, width+2);

    for (int i = 0; i < height; i++) {
        int j = 0;
#ifdef ENABLE_NEON
        uint8x8_t v_tone = vcreate_u8((i % 2) == 0 ? 0xff00ff00ff00ff00 : 0x00ff00ff00ff00ff);
        for ( ; j < width-7; j+=8) {
            uint8x8_t v_src = vld1_u8(p_src+i*width+j);
            uint8x8_t v_mask_white = vcge_u8(v_src, vdup_n_u8(threshold_gray_and_white));
            uint8x8_t v_mask_gray = veor_u8(v_mask_white, vcge_u8(v_src, vdup_n_u8(threshold_black_and_gray)));
            uint8x8_t v_lum = vorr_u8(v_mask_white, vand_u8(v_tone, v_mask_gray));

            int16x8_t v_src00 = vreinterpretq_s16_u16(vmovl_u8(vld1_u8(line0+j+0)));
            int16x8_t v_src01 = vreinterpretq_s16_u16(vmovl_u8(vld1_u8(line0+j+1)));
            int16x8_t v_src02 = vreinterpretq_s16_u16(vmovl_u8(vld1_u8(line0+j+2)));
            int16x8_t v_src10 = vreinterpretq_s16_u16(vmovl_u8(vld1_u8(line1+j+0)));
            int16x8_t v_src12 = vreinterpretq_s16_u16(vmovl_u8(vld1_u8(line1+j+2)));
            int16x8_t v_src20 = vreinterpretq_s16_u16(vmovl_u8(vld1_u8(line2+j+0)));
            int16x8_t v_src21 = vreinterpretq_s16_u16(vmovl_u8(vld1_u8(line2+j+1)));
            int16x8_t v_src22 = vreinterpretq_s16_u16(vmovl_u8(vld1_u8(line2+j+2)));

            int16x8_t v_fx = vdupq_n_s16(0);
            v_fx = vsubq_s16(v_fx, v_src00);
            v_fx = vsubq_s16(v_fx, vshlq_n_s16(v_src10, 1));
            v_fx = vsubq_s16(v_fx, v_src20);
            v_fx = vaddq_s16(v_fx, v_src02);
            v_fx = vaddq_s16(v_fx, vshlq_n_s16(v_src12, 1));
            v_fx = vaddq_s16(v_fx, v_src22);

            int16x8_t v_fy = vdupq_n_s16(0);
            v_fy = vsubq_s16(v_fy, v_src00);
            v_fy = vsubq_s16(v_fy, vshlq_n_s16(v_src01, 1));
            v_fy = vsubq_s16(v_fy, v_src02);
            v_fy = vaddq_s16(v_fy, v_src20);
            v_fy = vaddq_s16(v_fy, vshlq_n_s16(v_src21, 1));
            v_fy = vaddq_s16(v_fy, v_src22);

            uint8x8_t v_edge = vclt_u8(vqmovun_s16(vaddq_s16(vabsq_s16(v_fx), vabsq_s16(v_fy))), vdup_n_u8(threshold_edge));
            v_lum = vand_u8(v_lum, v_edge);

            uint8x8x4_t v_dst;
            v_dst.val[0] = v_lum;
            v_dst.val[1] = v_lum;
            v_dst.val[2] = v_lum;
            v_dst.val[3] = vdup_n_u8(0xff);
            vst4_u8(p_dst+(i*width+j)*4, v_dst);
        }
#endif
        for ( ; j < width; j++) {
            lum = p_src[i*width+j];

            if (lum < threshold_black_and_gray) {
                lum = 0x00;
            }
            else if (lum < threshold_gray_and_white) {
                lum = ((i+j) % 2) == 0 ? 0x00 : 0xff;
            }
            else {
                lum = 0xff;
            }

            if (lum != 0x00) {
                fx = 0;
                fx += line0[j+0] * -1;
                fx += line1[j+0] * -2;
                fx += line2[j+0] * -1;
                fx += line0[j+2] *  1;
                fx += line1[j+2] *  2;
                fx += line2[j+2] *  1;
                fy = 0;
                fy += line0[j+0] * -1;
                fy += line0[j+1] * -2;
                fy += line0[j+2] * -1;
                fy += line2[j+0] *  1;
                fy += line2[j+1] *  2;
                fy += line2[j+2] *  1;
                lum = abs(fx) + abs(fy);
                lum = lum < threshold_edge ? 0xff : 0x00;
            }

            p_dst[(i*width+j)*4+0] = p_dst[(i*width+j)*4+1] = p_dst[(i*width+j)*4+2] = lum;
            p_dst[(i*width+j)*4+3] = 0xff;
        }

        // update line buffers.
        unsigned char *tmp = line0;
        line0 = line1;
        line1 = line2;
        line2 = tmp;
        memcpy(line2+1, p_src+(i+1 > height-1 ? height-1 : i+1)*width, width);
        line2[0] = line2[1];
        line2[width+1] = line2[width];
    }

    env->ReleasePrimitiveArrayCritical(src, p_src, 0);
    AndroidBitmap_unlockPixels(env, dst);

    return true;
}
